@@ -36,7 +36,7 @@
 #include <babeltrace/list.h>
 #include <babeltrace/types.h>
 #include <babeltrace/ctf/metadata.h>
-#include <babeltrace/uuid.h>
+#include <babeltrace/compat/uuid.h>
 #include <babeltrace/endian.h>
 #include <babeltrace/ctf/events-internal.h>
 #include "ctf-scanner.h"
@@ -95,10 +95,11 @@ char *concatenate_unary_strings(struct bt_list_head *head)
 	bt_list_for_each_entry(node, head, siblings) {
 		char *src_string;
 
-		assert(node->type == NODE_UNARY_EXPRESSION);
-		assert(node->u.unary_expression.type == UNARY_STRING);
-		assert((node->u.unary_expression.link == UNARY_LINK_UNKNOWN)
-			^ (i != 0));
+		if (node->type != NODE_UNARY_EXPRESSION
+				|| node->u.unary_expression.type != UNARY_STRING
+				|| !((node->u.unary_expression.link != UNARY_LINK_UNKNOWN)
+					^ (i == 0)))
+			return NULL;
 		switch (node->u.unary_expression.link) {
 		case UNARY_DOTLINK:
 			g_string_append(str, ".");
@@ -129,10 +130,11 @@ GQuark get_map_clock_name_value(struct bt_list_head *head)
 	bt_list_for_each_entry(node, head, siblings) {
 		char *src_string;
 
-		assert(node->type == NODE_UNARY_EXPRESSION);
-		assert(node->u.unary_expression.type == UNARY_STRING);
-		assert((node->u.unary_expression.link == UNARY_LINK_UNKNOWN)
-			^ (i != 0));
+		if (node->type != NODE_UNARY_EXPRESSION
+			|| node->u.unary_expression.type != UNARY_STRING
+			|| !((node->u.unary_expression.link != UNARY_LINK_UNKNOWN)
+				^ (i == 0)))
+			return 0;
 		/* needs to be chained with . */
 		switch (node->u.unary_expression.link) {
 		case UNARY_DOTLINK:
@@ -184,10 +186,11 @@ int get_unary_unsigned(struct bt_list_head *head, uint64_t *value)
 	int i = 0;
 
 	bt_list_for_each_entry(node, head, siblings) {
-		assert(node->type == NODE_UNARY_EXPRESSION);
-		assert(node->u.unary_expression.type == UNARY_UNSIGNED_CONSTANT);
-		assert(node->u.unary_expression.link == UNARY_LINK_UNKNOWN);
-		assert(i == 0);
+		if (node->type != NODE_UNARY_EXPRESSION
+				|| node->u.unary_expression.type != UNARY_UNSIGNED_CONSTANT
+				|| node->u.unary_expression.link != UNARY_LINK_UNKNOWN
+				|| i != 0)
+			return -EINVAL;
 		*value = node->u.unary_expression.u.unsigned_constant;
 		i++;
 	}
@@ -215,11 +218,12 @@ int get_unary_signed(struct bt_list_head *head, int64_t *value)
 	int i = 0;
 
 	bt_list_for_each_entry(node, head, siblings) {
-		assert(node->type == NODE_UNARY_EXPRESSION);
-		assert(node->u.unary_expression.type == UNARY_UNSIGNED_CONSTANT
-			|| node->u.unary_expression.type == UNARY_SIGNED_CONSTANT);
-		assert(node->u.unary_expression.link == UNARY_LINK_UNKNOWN);
-		assert(i == 0);
+		if (node->type != NODE_UNARY_EXPRESSION
+				|| node->u.unary_expression.type != UNARY_UNSIGNED_CONSTANT
+				|| (node->u.unary_expression.type != UNARY_UNSIGNED_CONSTANT && node->u.unary_expression.type != UNARY_SIGNED_CONSTANT)
+				|| node->u.unary_expression.link != UNARY_LINK_UNKNOWN
+				|| i != 0)
+			return -EINVAL;
 		switch (node->u.unary_expression.type) {
 		case UNARY_UNSIGNED_CONSTANT:
 			*value = (int64_t) node->u.unary_expression.u.unsigned_constant;
@@ -228,7 +232,7 @@ int get_unary_signed(struct bt_list_head *head, int64_t *value)
 			*value = node->u.unary_expression.u.signed_constant;
 			break;
 		default:
-			assert(0);
+			return -EINVAL;
 		}
 		i++;
 	}
@@ -245,10 +249,11 @@ int get_unary_uuid(struct bt_list_head *head, unsigned char *uuid)
 	bt_list_for_each_entry(node, head, siblings) {
 		const char *src_string;
 
-		assert(node->type == NODE_UNARY_EXPRESSION);
-		assert(node->u.unary_expression.type == UNARY_STRING);
-		assert(node->u.unary_expression.link == UNARY_LINK_UNKNOWN);
-		assert(i == 0);
+		if (node->type != NODE_UNARY_EXPRESSION
+				|| node->u.unary_expression.type != UNARY_STRING
+				|| node->u.unary_expression.link != UNARY_LINK_UNKNOWN
+				|| i != 0)
+			return -EINVAL;
 		src_string = node->u.unary_expression.u.string;
 		ret = babeltrace_uuid_parse(src_string, uuid);
 	}
@@ -263,16 +268,25 @@ struct ctf_stream_declaration *trace_stream_lookup(struct ctf_trace *trace, uint
 	return g_ptr_array_index(trace->streams, stream_id);
 }
 
+static
+struct ctf_event_declaration *stream_event_lookup(struct ctf_stream_declaration *stream, uint64_t event_id)
+{
+	if (stream->events_by_id->len <= event_id)
+		return NULL;
+	return g_ptr_array_index(stream->events_by_id, event_id);
+}
+
 static
 struct ctf_clock *trace_clock_lookup(struct ctf_trace *trace, GQuark clock_name)
 {
-	return g_hash_table_lookup(trace->clocks, (gpointer) (unsigned long) clock_name);
+	return g_hash_table_lookup(trace->parent.clocks, (gpointer) (unsigned long) clock_name);
 }
 
 static
 int visit_type_specifier(FILE *fd, struct ctf_node *type_specifier, GString *str)
 {
-	assert(type_specifier->type == NODE_TYPE_SPECIFIER);
+	if (type_specifier->type != NODE_TYPE_SPECIFIER)
+		return -EINVAL;
 
 	switch (type_specifier->u.type_specifier.type) {
 	case TYPESPEC_VOID:
@@ -426,7 +440,9 @@ struct bt_declaration *ctf_type_declarator_visit(FILE *fd, int depth,
 	 */
 
 	if (node_type_declarator) {
-		assert(node_type_declarator->u.type_declarator.type != TYPEDEC_UNKNOWN);
+		if (node_type_declarator->u.type_declarator.type == TYPEDEC_UNKNOWN) {
+			return NULL;
+		}
 
 		/* TODO: gcc bitfields not supported yet. */
 		if (node_type_declarator->u.type_declarator.bitfield_len != NULL) {
@@ -500,7 +516,9 @@ struct bt_declaration *ctf_type_declarator_visit(FILE *fd, int depth,
 		}
 		first = _bt_list_first_entry(&node_type_declarator->u.type_declarator.u.nested.length, 
 				struct ctf_node, siblings);
-		assert(first->type == NODE_UNARY_EXPRESSION);
+		if (first->type != NODE_UNARY_EXPRESSION) {
+			return NULL;
+		}
 
 		switch (first->u.unary_expression.type) {
 		case UNARY_UNSIGNED_CONSTANT:
@@ -526,6 +544,8 @@ struct bt_declaration *ctf_type_declarator_visit(FILE *fd, int depth,
 			char *length_name = concatenate_unary_strings(&node_type_declarator->u.type_declarator.u.nested.length);
 			struct declaration_sequence *sequence_declaration;
 
+			if (!length_name)
+				return NULL;
 			sequence_declaration = bt_sequence_declaration_new(length_name, nested_declaration, declaration_scope);
 			if (!sequence_declaration) {
 				fprintf(fd, "[error] %s: cannot create sequence declaration.\n", __func__);
@@ -538,7 +558,7 @@ struct bt_declaration *ctf_type_declarator_visit(FILE *fd, int depth,
 			break;
 		}
 		default:
-			assert(0);
+			return NULL;
 		}
 
 		/* Pass it as content of outer container */
@@ -637,7 +657,7 @@ int ctf_typedef_visit(FILE *fd, int depth, struct declaration_scope *scope,
 	bt_list_for_each_entry(iter, type_declarators, siblings) {
 		struct bt_declaration *type_declaration;
 		int ret;
-	
+
 		type_declaration = ctf_type_declarator_visit(fd, depth,
 					type_specifier_list,
 					&identifier, iter,
@@ -773,7 +793,7 @@ int ctf_struct_declaration_list_visit(FILE *fd, int depth,
 		break;
 	default:
 		fprintf(fd, "[error] %s: unexpected node type %d\n", __func__, (int) iter->type);
-		assert(0);
+		return -EINVAL;
 	}
 	return 0;
 }
@@ -817,7 +837,7 @@ int ctf_variant_declaration_list_visit(FILE *fd, int depth,
 		break;
 	default:
 		fprintf(fd, "[error] %s: unexpected node type %d\n", __func__, (int) iter->type);
-		assert(0);
+		return -EINVAL;
 	}
 	return 0;
 }
@@ -838,7 +858,8 @@ struct bt_declaration *ctf_declaration_struct_visit(FILE *fd,
 	 * declaration: ref is only taken upon definition.
 	 */
 	if (!has_body) {
-		assert(name);
+		if (!name)
+			return NULL;
 		struct_declaration =
 			bt_lookup_struct_declaration(g_quark_from_string(name),
 						  declaration_scope);
@@ -852,7 +873,6 @@ struct bt_declaration *ctf_declaration_struct_visit(FILE *fd,
 		if (name) {
 			if (bt_lookup_struct_declaration(g_quark_from_string(name),
 						      declaration_scope)) {
-				
 				fprintf(fd, "[error] %s: struct %s already declared in scope\n", __func__, name);
 				return NULL;
 			}
@@ -863,7 +883,6 @@ struct bt_declaration *ctf_declaration_struct_visit(FILE *fd,
 			ret = get_unary_unsigned(min_align, &min_align_value);
 			if (ret) {
 				fprintf(fd, "[error] %s: unexpected unary expression for structure \"align\" attribute\n", __func__);
-				ret = -EINVAL;
 				goto error;
 			}
 		}
@@ -883,7 +902,8 @@ struct bt_declaration *ctf_declaration_struct_visit(FILE *fd,
 			ret = bt_register_struct_declaration(g_quark_from_string(name),
 					struct_declaration,
 					declaration_scope);
-			assert(!ret);
+			if (ret)
+				return NULL;
 		}
 		return &struct_declaration->p;
 	}
@@ -910,7 +930,8 @@ struct bt_declaration *ctf_declaration_variant_visit(FILE *fd,
 	 * declaration: ref is only taken upon definition.
 	 */
 	if (!has_body) {
-		assert(name);
+		if (!name)
+			return NULL;
 		untagged_variant_declaration =
 			bt_lookup_variant_declaration(g_quark_from_string(name),
 						   declaration_scope);
@@ -921,7 +942,6 @@ struct bt_declaration *ctf_declaration_variant_visit(FILE *fd,
 		if (name) {
 			if (bt_lookup_variant_declaration(g_quark_from_string(name),
 						       declaration_scope)) {
-				
 				fprintf(fd, "[error] %s: variant %s already declared in scope\n", __func__, name);
 				return NULL;
 			}
@@ -941,7 +961,8 @@ struct bt_declaration *ctf_declaration_variant_visit(FILE *fd,
 			ret = bt_register_variant_declaration(g_quark_from_string(name),
 					untagged_variant_declaration,
 					declaration_scope);
-			assert(!ret);
+			if (ret)
+				return NULL;
 		}
 	}
 	/*
@@ -979,7 +1000,8 @@ int ctf_enumerator_list_visit(FILE *fd, int depth,
 		bt_list_for_each_entry(iter, &enumerator->u.enumerator.values, siblings) {
 			int64_t *target;
 
-			assert(iter->type == NODE_UNARY_EXPRESSION);
+			if (iter->type != NODE_UNARY_EXPRESSION)
+				return -EINVAL;
 			if (nr_vals == 0)
 				target = &start;
 			else
@@ -1015,7 +1037,8 @@ int ctf_enumerator_list_visit(FILE *fd, int depth,
 		bt_list_for_each_entry(iter, &enumerator->u.enumerator.values, siblings) {
 			uint64_t *target;
 
-			assert(iter->type == NODE_UNARY_EXPRESSION);
+			if (iter->type != NODE_UNARY_EXPRESSION)
+				return -EINVAL;
 			if (nr_vals == 0)
 				target = &start;
 			else
@@ -1074,7 +1097,8 @@ struct bt_declaration *ctf_declaration_enum_visit(FILE *fd, int depth,
 	 * declaration: ref is only taken upon definition.
 	 */
 	if (!has_body) {
-		assert(name);
+		if (!name)
+			return NULL;
 		enum_declaration =
 			bt_lookup_enum_declaration(g_quark_from_string(name),
 						declaration_scope);
@@ -1086,7 +1110,6 @@ struct bt_declaration *ctf_declaration_enum_visit(FILE *fd, int depth,
 		if (name) {
 			if (bt_lookup_enum_declaration(g_quark_from_string(name),
 						    declaration_scope)) {
-				
 				fprintf(fd, "[error] %s: enum %s already declared in scope\n", __func__, name);
 				return NULL;
 			}
@@ -1135,7 +1158,8 @@ struct bt_declaration *ctf_declaration_enum_visit(FILE *fd, int depth,
 			ret = bt_register_enum_declaration(g_quark_from_string(name),
 					enum_declaration,
 					declaration_scope);
-			assert(!ret);
+			if (ret)
+				return NULL;
 			bt_declaration_unref(&enum_declaration->p);
 		}
 		return &enum_declaration->p;
@@ -1158,12 +1182,16 @@ struct bt_declaration *ctf_declaration_type_specifier_visit(FILE *fd, int depth,
 
 	str = g_string_new("");
 	ret = visit_type_specifier_list(fd, type_specifier_list, str);
-	if (ret)
+	if (ret) {
+		(void) g_string_free(str, TRUE);
 		return NULL;
+	}
 	str_c = g_string_free(str, FALSE);
 	id_q = g_quark_from_string(str_c);
 	g_free(str_c);
 	declaration = bt_lookup_declaration(id_q, declaration_scope);
+	if (!declaration)
+		return NULL;
 	bt_declaration_ref(declaration);
 	return declaration;
 }
@@ -1209,7 +1237,7 @@ int get_boolean(FILE *fd, int depth, struct ctf_node *unary_expression)
 		fprintf(fd, "[error] %s: unexpected unary expression type\n",
 			__func__);
 		return -EINVAL;
-	} 
+	}
 
 }
 
@@ -1282,7 +1310,8 @@ struct bt_declaration *ctf_declaration_integer_visit(FILE *fd, int depth,
 
 		left = _bt_list_first_entry(&expression->u.ctf_expression.left, struct ctf_node, siblings);
 		right = _bt_list_first_entry(&expression->u.ctf_expression.right, struct ctf_node, siblings);
-		assert(left->u.unary_expression.type == UNARY_STRING);
+		if (left->u.unary_expression.type != UNARY_STRING)
+			return NULL;
 		if (!strcmp(left->u.unary_expression.u.string, "signed")) {
 			signedness = get_boolean(fd, depth, right);
 			if (signedness < 0)
@@ -1298,6 +1327,11 @@ struct bt_declaration *ctf_declaration_integer_visit(FILE *fd, int depth,
 				return NULL;
 			}
 			size = right->u.unary_expression.u.unsigned_constant;
+			if (!size) {
+				fprintf(fd, "[error] %s: integer size: expecting non-zero constant\n",
+					__func__);
+				return NULL;
+			}
 			has_size = 1;
 		} else if (!strcmp(left->u.unary_expression.u.string, "align")) {
 			if (right->u.unary_expression.type != UNARY_UNSIGNED_CONSTANT) {
@@ -1454,9 +1488,9 @@ struct bt_declaration *ctf_declaration_floating_point_visit(FILE *fd, int depth,
 		struct ctf_trace *trace)
 {
 	struct ctf_node *expression;
-	uint64_t alignment = 1, exp_dig = 0, mant_dig = 0,
-		byte_order = trace->byte_order;
-	int has_alignment = 0, has_exp_dig = 0, has_mant_dig = 0;
+	uint64_t alignment = 1, exp_dig = 0, mant_dig = 0;
+	int byte_order = trace->byte_order, has_alignment = 0,
+		has_exp_dig = 0, has_mant_dig = 0;
 	struct declaration_float *float_declaration;
 
 	bt_list_for_each_entry(expression, expressions, siblings) {
@@ -1464,7 +1498,8 @@ struct bt_declaration *ctf_declaration_floating_point_visit(FILE *fd, int depth,
 
 		left = _bt_list_first_entry(&expression->u.ctf_expression.left, struct ctf_node, siblings);
 		right = _bt_list_first_entry(&expression->u.ctf_expression.right, struct ctf_node, siblings);
-		assert(left->u.unary_expression.type == UNARY_STRING);
+		if (left->u.unary_expression.type != UNARY_STRING)
+			return NULL;
 		if (!strcmp(left->u.unary_expression.u.string, "byte_order")) {
 			byte_order = get_byte_order(fd, depth, right, trace);
 			if (byte_order < 0)
@@ -1542,7 +1577,8 @@ struct bt_declaration *ctf_declaration_string_visit(FILE *fd, int depth,
 
 		left = _bt_list_first_entry(&expression->u.ctf_expression.left, struct ctf_node, siblings);
 		right = _bt_list_first_entry(&expression->u.ctf_expression.right, struct ctf_node, siblings);
-		assert(left->u.unary_expression.type == UNARY_STRING);
+		if (left->u.unary_expression.type != UNARY_STRING)
+			return NULL;
 		if (!strcmp(left->u.unary_expression.u.string, "encoding")) {
 			if (right->u.unary_expression.type != UNARY_STRING) {
 				fprintf(fd, "[error] %s: encoding: expecting string\n",
@@ -1572,11 +1608,13 @@ struct bt_declaration *ctf_type_specifier_list_visit(FILE *fd,
 	struct ctf_node *first;
 	struct ctf_node *node;
 
-	assert(type_specifier_list->type == NODE_TYPE_SPECIFIER_LIST);
+	if (type_specifier_list->type != NODE_TYPE_SPECIFIER_LIST)
+		return NULL;
 
 	first = _bt_list_first_entry(&type_specifier_list->u.type_specifier_list.head, struct ctf_node, siblings);
 
-	assert(first->type == NODE_TYPE_SPECIFIER);
+	if (first->type != NODE_TYPE_SPECIFIER)
+		return NULL;
 
 	node = first->u.type_specifier.node;
 
@@ -1665,6 +1703,8 @@ int ctf_event_declaration_visit(FILE *fd, int depth, struct ctf_node *node, stru
 		char *left;
 
 		left = concatenate_unary_strings(&node->u.ctf_expression.left);
+		if (!left)
+			return -EINVAL;
 		if (!strcmp(left, "name")) {
 			char *right;
 
@@ -1814,6 +1854,10 @@ int ctf_event_visit(FILE *fd, int depth, struct ctf_node *node,
 	struct ctf_event_declaration *event;
 	struct bt_ctf_event_decl *event_decl;
 
+	if (node->visited)
+		return 0;
+	node->visited = 1;
+
 	event_decl = g_new0(struct bt_ctf_event_decl, 1);
 	event = &event_decl->parent;
 	event->declaration_scope = bt_new_declaration_scope(parent_declaration_scope);
@@ -1853,6 +1897,13 @@ int ctf_event_visit(FILE *fd, int depth, struct ctf_node *node,
 		fprintf(fd, "[error] %s: missing id field in event declaration\n", __func__);
 		goto error;
 	}
+	/* Disallow re-using the same event ID in the same stream */
+	if (stream_event_lookup(event->stream, event->id)) {
+		ret = -EPERM;
+		fprintf(fd, "[error] %s: event ID %" PRIu64 " used more than once in stream %" PRIu64 "\n",
+			__func__, event->id, event->stream_id);
+		goto error;
+	}
 	if (event->stream->events_by_id->len <= event->id)
 		g_ptr_array_set_size(event->stream->events_by_id, event->id + 1);
 	g_ptr_array_index(event->stream->events_by_id, event->id) = event;
@@ -1901,6 +1952,8 @@ int ctf_stream_declaration_visit(FILE *fd, int depth, struct ctf_node *node, str
 		char *left;
 
 		left = concatenate_unary_strings(&node->u.ctf_expression.left);
+		if (!left)
+			return -EINVAL;
 		if (!strcmp(left, "id")) {
 			if (CTF_STREAM_FIELD_IS_SET(stream, stream_id)) {
 				fprintf(fd, "[error] %s: id already declared in stream declaration\n", __func__);
@@ -2002,6 +2055,12 @@ int ctf_stream_visit(FILE *fd, int depth, struct ctf_node *node,
 	struct ctf_node *iter;
 	struct ctf_stream_declaration *stream;
 
+	if (node) {
+		if (node->visited)
+			return 0;
+		node->visited = 1;
+	}
+
 	stream = g_new0(struct ctf_stream_declaration, 1);
 	stream->declaration_scope = bt_new_declaration_scope(parent_declaration_scope);
 	stream->events_by_id = g_ptr_array_new();
@@ -2081,6 +2140,8 @@ int ctf_trace_declaration_visit(FILE *fd, int depth, struct ctf_node *node, stru
 		char *left;
 
 		left = concatenate_unary_strings(&node->u.ctf_expression.left);
+		if (!left)
+			return -EINVAL;
 		if (!strcmp(left, "major")) {
 			if (CTF_TRACE_FIELD_IS_SET(trace, major)) {
 				fprintf(fd, "[error] %s: major already declared in trace declaration\n", __func__);
@@ -2131,8 +2192,10 @@ int ctf_trace_declaration_visit(FILE *fd, int depth, struct ctf_node *node, stru
 
 			right = _bt_list_first_entry(&node->u.ctf_expression.right, struct ctf_node, siblings);
 			byte_order = get_trace_byte_order(fd, depth, right);
-			if (byte_order < 0)
-				return -EINVAL;
+			if (byte_order < 0) {
+				ret = -EINVAL;
+				goto error;
+			}
 
 			if (CTF_TRACE_FIELD_IS_SET(trace, byte_order)
 				&& byte_order != trace->byte_order) {
@@ -2197,8 +2260,13 @@ int ctf_trace_visit(FILE *fd, int depth, struct ctf_node *node, struct ctf_trace
 	int ret = 0;
 	struct ctf_node *iter;
 
+	if (!trace->restart_root_decl && node->visited)
+		return 0;
+	node->visited = 1;
+
 	if (trace->declaration_scope)
 		return -EEXIST;
+
 	trace->declaration_scope = bt_new_declaration_scope(trace->root_declaration_scope);
 	trace->streams = g_ptr_array_new();
 	trace->event_declarations = g_ptr_array_new();
@@ -2258,6 +2326,8 @@ int ctf_clock_declaration_visit(FILE *fd, int depth, struct ctf_node *node,
 		char *left;
 
 		left = concatenate_unary_strings(&node->u.ctf_expression.left);
+		if (!left)
+			return -EINVAL;
 		if (!strcmp(left, "name")) {
 			char *right;
 
@@ -2364,6 +2434,7 @@ int ctf_clock_declaration_visit(FILE *fd, int depth, struct ctf_node *node,
 				goto error;
 			}
 			clock->absolute = ret;
+			ret = 0;
 		} else {
 			fprintf(fd, "[warning] %s: attribute \"%s\" is unknown in clock declaration.\n", __func__, left);
 		}
@@ -2387,6 +2458,10 @@ int ctf_clock_visit(FILE *fd, int depth, struct ctf_node *node, struct ctf_trace
 	struct ctf_node *iter;
 	struct ctf_clock *clock;
 
+	if (node->visited)
+		return 0;
+	node->visited = 1;
+
 	clock = g_new0(struct ctf_clock, 1);
 	/* Default clock frequency is set to 1000000000 */
 	clock->freq = 1000000000ULL;
@@ -2411,13 +2486,13 @@ int ctf_clock_visit(FILE *fd, int depth, struct ctf_node *node, struct ctf_trace
 		fprintf(fd, "[error] %s: missing name field in clock declaration\n", __func__);
 		goto error;
 	}
-	if (g_hash_table_size(trace->clocks) > 0) {
+	if (g_hash_table_size(trace->parent.clocks) > 0) {
 		fprintf(fd, "[error] Only CTF traces with a single clock description are supported by this babeltrace version.\n");
 		ret = -EINVAL;
 		goto error;
 	}
-	trace->single_clock = clock;
-	g_hash_table_insert(trace->clocks, (gpointer) (unsigned long) clock->name, clock);
+	trace->parent.single_clock = clock;
+	g_hash_table_insert(trace->parent.clocks, (gpointer) (unsigned long) clock->name, clock);
 	return 0;
 
 error:
@@ -2451,8 +2526,8 @@ void ctf_clock_default(FILE *fd, int depth, struct ctf_trace *trace)
 		clock->absolute = 0;	/* Not an absolute reference across traces */
 	}
 
-	trace->single_clock = clock;
-	g_hash_table_insert(trace->clocks, (gpointer) (unsigned long) clock->name, clock);
+	trace->parent.single_clock = clock;
+	g_hash_table_insert(trace->parent.clocks, (gpointer) (unsigned long) clock->name, clock);
 }
 
 static
@@ -2476,6 +2551,8 @@ int ctf_callsite_declaration_visit(FILE *fd, int depth, struct ctf_node *node,
 		char *left;
 
 		left = concatenate_unary_strings(&node->u.ctf_expression.left);
+		if (!left)
+			return -EINVAL;
 		if (!strcmp(left, "name")) {
 			char *right;
 
@@ -2575,6 +2652,10 @@ int ctf_callsite_visit(FILE *fd, int depth, struct ctf_node *node, struct ctf_tr
 	struct ctf_callsite *callsite;
 	struct ctf_callsite_dups *cs_dups;
 
+	if (node->visited)
+		return 0;
+	node->visited = 1;
+
 	callsite = g_new0(struct ctf_callsite, 1);
 	bt_list_for_each_entry(iter, &node->u.callsite.declaration_list, siblings) {
 		ret = ctf_callsite_declaration_visit(fd, depth + 1, iter, callsite, trace);
@@ -2647,6 +2728,8 @@ int ctf_env_declaration_visit(FILE *fd, int depth, struct ctf_node *node,
 		char *left;
 
 		left = concatenate_unary_strings(&node->u.ctf_expression.left);
+		if (!left)
+			return -EINVAL;
 		if (!strcmp(left, "vpid")) {
 			uint64_t v;
 
@@ -2762,6 +2845,11 @@ int ctf_env_declaration_visit(FILE *fd, int depth, struct ctf_node *node,
 				char *right;
 
 				right = concatenate_unary_strings(&node->u.ctf_expression.right);
+				if (!right) {
+					fprintf(fd, "[warning] %s: unexpected unary expression for env\n", __func__);
+					ret = -EINVAL;
+					goto error;
+				}
 				printf_verbose("env.%s = \"%s\"\n", left, right);
 				g_free(right);
 			} else if (is_unary_unsigned(&node->u.ctf_expression.right)) {
@@ -2769,14 +2857,16 @@ int ctf_env_declaration_visit(FILE *fd, int depth, struct ctf_node *node,
 				int ret;
 
 				ret = get_unary_unsigned(&node->u.ctf_expression.right, &v);
-				assert(ret == 0);
+				if (ret)
+					goto error;
 				printf_verbose("env.%s = %" PRIu64 "\n", left, v);
 			} else if (is_unary_signed(&node->u.ctf_expression.right)) {
 				int64_t v;
 				int ret;
 
 				ret = get_unary_signed(&node->u.ctf_expression.right, &v);
-				assert(ret == 0);
+				if (ret)
+					goto error;
 				printf_verbose("env.%s = %" PRId64 "\n", left, v);
 			} else {
 				printf_verbose("%s: attribute \"%s\" has unknown type.\n", __func__, left);
@@ -2800,6 +2890,10 @@ int ctf_env_visit(FILE *fd, int depth, struct ctf_node *node, struct ctf_trace *
 	int ret = 0;
 	struct ctf_node *iter;
 
+	if (node->visited)
+		return 0;
+	node->visited = 1;
+
 	trace->env.vpid = -1;
 	trace->env.procname[0] = '\0';
 	trace->env.hostname[0] = '\0';
@@ -2821,6 +2915,10 @@ int ctf_root_declaration_visit(FILE *fd, int depth, struct ctf_node *node, struc
 {
 	int ret = 0;
 
+	if (!trace->restart_root_decl && node->visited)
+		return 0;
+	node->visited = 1;
+
 	switch (node->type) {
 	case NODE_TYPEDEF:
 		ret = ctf_typedef_visit(fd, depth + 1,
@@ -2866,12 +2964,11 @@ int ctf_visitor_construct_metadata(FILE *fd, int depth, struct ctf_node *node,
 {
 	int ret = 0;
 	struct ctf_node *iter;
-	int env_clock_done = 0;
 
 	printf_verbose("CTF visitor: metadata construction...\n");
 	trace->byte_order = byte_order;
-	trace->clocks = g_hash_table_new_full(g_direct_hash, g_direct_equal,
-				NULL, clock_free);
+	trace->parent.clocks = g_hash_table_new_full(g_direct_hash,
+				g_direct_equal, NULL, clock_free);
 	trace->callsites = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 				NULL, callsite_free);
 
@@ -2880,24 +2977,21 @@ retry:
 
 	switch (node->type) {
 	case NODE_ROOT:
-		if (!env_clock_done) {
-			/*
-			 * declarations need to query clock hash table,
-			 * so clock need to be treated first.
-			 */
-			if (bt_list_empty(&node->u.root.clock)) {
-				ctf_clock_default(fd, depth + 1, trace);
-			} else {
-				bt_list_for_each_entry(iter, &node->u.root.clock, siblings) {
-					ret = ctf_clock_visit(fd, depth + 1, iter,
-							      trace);
-					if (ret) {
-						fprintf(fd, "[error] %s: clock declaration error\n", __func__);
-						goto error;
-					}
+		/*
+		 * declarations need to query clock hash table,
+		 * so clock need to be treated first.
+		 */
+		if (bt_list_empty(&node->u.root.clock)) {
+			ctf_clock_default(fd, depth + 1, trace);
+		} else {
+			bt_list_for_each_entry(iter, &node->u.root.clock, siblings) {
+				ret = ctf_clock_visit(fd, depth + 1, iter,
+						      trace);
+				if (ret) {
+					fprintf(fd, "[error] %s: clock declaration error\n", __func__);
+					goto error;
 				}
 			}
-			env_clock_done = 1;
 		}
 		bt_list_for_each_entry(iter, &node->u.root.declaration_list,
 					siblings) {
@@ -2910,6 +3004,7 @@ retry:
 		bt_list_for_each_entry(iter, &node->u.root.trace, siblings) {
 			ret = ctf_trace_visit(fd, depth + 1, iter, trace);
 			if (ret == -EINTR) {
+				trace->restart_root_decl = 1;
 				bt_free_declaration_scope(trace->root_declaration_scope);
 				/*
 				 * Need to restart creation of type
@@ -2923,6 +3018,7 @@ retry:
 				goto error;
 			}
 		}
+		trace->restart_root_decl = 0;
 		bt_list_for_each_entry(iter, &node->u.root.callsite, siblings) {
 			ret = ctf_callsite_visit(fd, depth + 1, iter,
 					      trace);
@@ -2973,7 +3069,7 @@ retry:
 error:
 	bt_free_declaration_scope(trace->root_declaration_scope);
 	g_hash_table_destroy(trace->callsites);
-	g_hash_table_destroy(trace->clocks);
+	g_hash_table_destroy(trace->parent.clocks);
 	return ret;
 }
 
@@ -3072,7 +3168,7 @@ int ctf_destroy_metadata(struct ctf_trace *trace)
 	bt_free_declaration_scope(trace->declaration_scope);
 
 	g_hash_table_destroy(trace->callsites);
-	g_hash_table_destroy(trace->clocks);
+	g_hash_table_destroy(trace->parent.clocks);
 
 	metadata_stream = container_of(trace->metadata, struct ctf_file_stream, parent);
 	g_free(metadata_stream);
